use crate::{
    constant::*,
    util::*,
};

#[derive(Clone, Copy)]
pub struct MoveSet {
    pub direction:u32,
    pub stride:u32,
    pub alt:Option<u32>,
}
impl MoveSet {
    pub const fn new() -> Self
    {
        Self {
            direction:0,
            stride:0,
            alt:None,
        }
    }

    pub const fn add(mut self, bits:u32) -> Self
    {
        self.direction |= bits;
        self
    }

    pub const fn add_stride(mut self, bits:u32, mode:u32) -> Self
    {
        self.direction |= bits;
        
        let mut bits_set = bits;
        while bits_set != 0 {
            let mask = lsb(bits_set);
            let index = ffs(mask);

            self.stride |= (mode & 3) << (index * 2);

            bits_set &= !mask;
        }

        self
    }

    pub const fn add_alt(mut self, mode:u32) -> Self
    {
        self.alt = Some(mode);
        self
    }

    pub fn rotate(&self) -> Self
    {
        Self {
            direction:rotate6(self.direction),
            stride:rotate12(self.stride),
            alt:None,
        }
    }
}

#[derive(Clone, Copy)]
pub struct PieceClass {
    pub name:&'static str,
    pub moves:MoveSet,
    pub pmoves:MoveSet,
}

pub const PIECES :[PieceClass; PIECES_COUNT] = [
    PieceClass {
        name: "Militia",
        moves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(5)),
        pmoves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5)),
    },
    PieceClass {
        name: "Lance",
        moves: MoveSet::new()
            .add(bit(1) | bit(5))
            .add_stride(bit(0), 3),
        pmoves: MoveSet::new()
            .add(bit(0) | bit(3))
            .add_stride(bit(1) | bit(2) | bit(4) | bit(5), 1),
    },
    PieceClass {
        name: "Knight",
        moves: MoveSet::new()
            .add(bit(6) | bit(8) | bit(9) | bit(11) | bit(13) | bit(14) | bit(16) | bit(17)),
        pmoves: MoveSet::new()
            .add(bit(6) | bit(8) | bit(9) | bit(11) | bit(13) | bit(14) | bit(16) | bit(17))
            .add_alt(1),
    },
    PieceClass {
        name: "Tower",
        moves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(3) | bit(5) | bit(6) | bit(11)),
        pmoves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5) | bit(6) | bit(8) | bit(9) | bit(11)),
    },
    PieceClass {
        name: "Castle",
        moves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(2) | bit(4) | bit(5) | bit(7) | bit(10)),
        pmoves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5) | bit(7) | bit(10))
            .add_alt(2),
    },
    PieceClass {
        name: "Dragon",
        moves: MoveSet::new()
            .add_stride(bit(6) | bit(7) | bit(8) | bit(9) | bit(10) | bit(11), 3),
        pmoves: MoveSet::new()
            .add_stride(bit(6) | bit(7) | bit(8) | bit(9) | bit(10) | bit(11), 3)
            .add(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5)),
    },
    PieceClass {
        name: "Behemoth",
        moves: MoveSet::new()
            .add_stride(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5), 3),
        pmoves: MoveSet::new()
            .add_stride(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5), 3)
            .add(bit(12) | bit(13) | bit(14) | bit(15) | bit(16) | bit(17)),
    },
    PieceClass {
        name: "Heart",
        moves: MoveSet::new()
            .add(bit(0) | bit(1) | bit(2) | bit(3) | bit(4) | bit(5) | bit(7) | bit(10)),
        pmoves: MoveSet::new(),
    },
];

#[derive(Clone, Copy)]
pub struct Piece {
    pub class:u8,
    pub player:u8,
    pub promoted:bool,
    pub tile:u8,
    pub blocking:u32,
}
impl Piece {
    pub fn new(class:u8, player:u8) -> Self
    {
        Self {
            class,
            player,
            promoted:false,
            tile:0,
            blocking:0,
        }
    }

    pub fn at(mut self, tile:u8) -> Self
    {
        self.tile = tile;
        self
    }

    pub fn promote(mut self, promoted:bool) -> Self
    {
        self.promoted = promoted;
        self
    }

    pub fn moves(&self) -> MoveSet
    {
        let class = &crate::piece::PIECES[self.class as usize];
        
        let moves = if self.promoted {
            class.pmoves
        } else {
            class.moves
        };

        if self.player == 0 {
            moves
        } else {
            moves.rotate()
        }
    }

    pub fn has_promotion(&self) -> bool
    {
        !self.promoted && PIECES[self.class as usize].pmoves.direction != 0
    }
}
