use super::*;

use std::{
    net::SocketAddr, sync::Arc
};
use tokio::{net::{TcpListener, TcpStream}, sync::RwLock, task::JoinHandle};
use rustls::{
    pki_types::{CertificateDer, PrivateKeyDer}, server::Acceptor, ServerConfig
};
use tokio_rustls::{server, LazyConfigAcceptor};

use certstore::CertificateStore;

struct PKP {
    pub certs:Vec<CertificateDer<'static>>,
    pub key:PrivateKeyDer<'static>,
}

type CertStorePtr = Arc<RwLock<CertificateStore>>;

pub struct TlsStream {
    stream:server::TlsStream<TcpStream>,
}
impl TlsStream { }
impl Stream for TlsStream {
    type Socket = server::TlsStream<TcpStream>;

    fn recv(&mut self, _buffer:&mut [u8]) -> Result<usize, ()> {
        //match self.stream.read(&mut buffer) {
        //    Ok(size) => Ok(size),
        //    Err(_) => Err(())
        //}
        Err(())
    }

    fn send(&mut self, _buffer:&mut [u8]) -> Result<(), ()> {
        Err(())
    }

    fn stream(&mut self) -> &mut Self::Socket {
        &mut self.stream
    }

    fn to_stream(self) -> Self::Socket {
        self.stream
    }
}

pub struct TlsServer {
    certificates:CertStorePtr,
    listener:Option<TcpListener>,
}
impl TlsServer {
pub fn new() -> Self
    {
        Self {
            certificates:Arc::new(RwLock::new(CertificateStore::new())),
            listener:None,
        }
    }

    pub async fn add_cert(&mut self, domain:&str, cert_file:&str, key_file:&str) -> Result<(),()>
    {
        let mut cert_handle = self.certificates.write().await;
        match cert_handle.add(domain, cert_file, key_file) {
            Ok(_) => Ok(()),
            Err(_) => Err(())
        }
    }

    pub async fn remove_cert(&mut self, domain:&str) -> Result<(),()>
    {
        let mut cert_handle = self.certificates.write().await;
        match cert_handle.remove(domain) {
            Ok(_) => Ok(()),
            Err(_) => Err(())
        }
    }

    async fn get_server_config(certificates:CertStorePtr, domain :&str) -> Option<Arc<ServerConfig>>
    {
        match certificates.read().await.get(domain) {
            Ok(config) => {
                let server_config = ServerConfig::builder()
                    .with_no_client_auth()
                    .with_single_cert(config.certs.clone(), config.key.clone_key());
                match server_config {
                    Ok(config) => Some(Arc::new(config)),
                    Err(_) => None
                }
            }
            Err(_) => None
        }
    }
}
impl Server for TlsServer {
    type Stream = TlsStream;

    async fn bind(&mut self, interface:&str) -> Result<(), ()> {
        match tokio::net::TcpListener::bind(interface).await {
            Ok(listener) => {
                self.listener = Some(listener);
                Ok(())
            }
            Err(_) => Err(())
        }
    }

    async fn accept<F, Fut, Args>(&self, callback:F, args:Args) -> Result<JoinHandle<Result<(),()>>,()>
        where
            F: Fn(Self::Stream, SocketAddr, Args) -> Fut + Send + Sync + 'static,
            Fut: std::future::Future<Output = Result<(),()>> + Send + 'static,
            Args: Send + 'static
    {
        match &self.listener {
            Some(listener) => {
                if let Ok((socket, client_addr)) = listener.accept().await {
                    let acceptor = LazyConfigAcceptor::new(Acceptor::default(), socket);
                    let certificates = self.certificates.clone();

                    Ok(tokio::spawn(async move {
                        match acceptor.await {
                            Ok(hs) => {
                                let hello = hs.client_hello();
                                if let Some(server_name) = hello.server_name() {
                                    match Self::get_server_config(certificates, server_name).await {
                                        Some(server_config) => {
                                            match hs.into_stream(server_config).await {
                                                Ok(tlsstream) => {
                                                    callback(Self::Stream { stream:tlsstream }, client_addr, args).await
                                                }
                                                Err(_) => Err(())
                                            }
                                        }
                                        None => Err(())
                                    }
                                }
                                else { Err(()) }
                            }
                            Err(e) => {
                                println!("[connection failure] {:}", e.to_string());
                                Err(())
                            }
                        }
                    }))
                }
                else { Err(()) }
            }
            None => Err(())
        }
    }

    fn close(&mut self) {
        self.listener = None;
    }
}
